/*
    MPlayer Winamp-Plugin Plugin
    Copyright (C) 2004-2005 Jrg Walter
    More Copyright: see below.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 This program was derived from:
   xmms-winamp
   plugin helper binary, interface to Winamp plugin
   
   (C) 2002 Ulrich Hecht
   Licensed under the terms of the GNU General Public License v2
 */

#include <sys/mman.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <libgen.h>

#include <windows.h>

#define OUT_VER 0x10

typedef struct 
{
	int version;				// module version (OUT_VER)
	char *description;			// description of module, with version string
	int id;						// module id. each input module gets its own. non-nullsoft modules should
								// be >= 65536. 

	HWND hMainWindow;			// winamp's main window (filled in by winamp)
	HINSTANCE hDllInstance;		// DLL instance handle (filled in by winamp)

	void (*Config)(HWND hwndParent); // configuration dialog 
	void (*About)(HWND hwndParent);  // about dialog

	void (*Init)();				// called when loaded
	void (*Quit)();				// called when unloaded

	int (*Open)(int samplerate, int numchannels, int bitspersamp, int bufferlenms, int prebufferms); 
					// returns >=0 on success, <0 on failure
					// NOTENOTENOTE: bufferlenms and prebufferms are ignored in most if not all output plug-ins. 
					//    ... so don't expect the max latency returned to be what you asked for.
					// returns max latency in ms (0 for diskwriters, etc)
					// bufferlenms and prebufferms must be in ms. 0 to use defaults. 
					// prebufferms must be <= bufferlenms

	void (*Close)();	// close the ol' output device.

	int (*Write)(char *buf, int len);	
					// 0 on success. Len == bytes to write (<= 8192 always). buf is straight audio data. 
					// 1 returns not able to write (yet). Non-blocking, always.

	int (*CanWrite)();	// returns number of bytes possible to write at a given time. 
						// Never will decrease unless you call Write (or Close, heh)

	int (*IsPlaying)(); // non0 if output is still going or if data in buffers waiting to be
						// written (i.e. closing while IsPlaying() returns 1 would truncate the song

	int (*Pause)(int pause); // returns previous pause state

	void (*SetVolume)(int volume); // volume is 0-255
	void (*SetPan)(int pan); // pan is -128 to 128

	void (*Flush)(int t);	// flushes buffers and restarts output at time t (in ms) 
							// (used for seeking)

	int (*GetOutputTime)(); // returns played time in MS
	int (*GetWrittenTime)(); // returns time written in MS (used for synching up vis stuff)

} Out_Module;

#define IN_VER 0x100

typedef struct 
{
	int version;				// module type (IN_VER)
	char *description;			// description of module, with version string

	HWND hMainWindow;			// winamp's main window (filled in by winamp)
	HINSTANCE hDllInstance;		// DLL instance handle (Also filled in by winamp)

	char *FileExtensions;		// "mp3\0Layer 3 MPEG\0mp2\0Layer 2 MPEG\0mpg\0Layer 1 MPEG\0"
								// May be altered from Config, so the user can select what they want
	
	int is_seekable;			// is this stream seekable? 
	int UsesOutputPlug;			// does this plug-in use the output plug-ins? (musn't ever change, ever :)

	void (*Config)(HWND hwndParent); // configuration dialog
	void (*About)(HWND hwndParent);  // about dialog

	void (*Init)();				// called at program init
	void (*Quit)();				// called at program quit

	void (*GetFileInfo)(char *file, char *title, int *length_in_ms); // if file == NULL, current playing is used
	int (*InfoBox)(char *file, HWND hwndParent);
	
	int (*IsOurFile)(char *fn);	// called before extension checks, to allow detection of mms://, etc
	// playback stuff
	int (*Play)(char *fn);		// return zero on success, -1 on file-not-found, some other value on other (stopping winamp) error
	void (*Pause)();			// pause stream
	void (*UnPause)();			// unpause stream
	int (*IsPaused)();			// ispaused? return 1 if paused, 0 if not
	void (*Stop)();				// stop (unload) stream

	// time stuff
	int (*GetLength)();			// get length in ms
	int (*GetOutputTime)();		// returns current output time in ms. (usually returns outMod->GetOutputTime()
	void (*SetOutputTime)(int time_in_ms);	// seeks to point in stream (in ms). Usually you signal yoru thread to seek, which seeks and calls outMod->Flush()..

	// volume stuff
	void (*SetVolume)(int volume);	// from 0 to 255.. usually just call outMod->SetVolume
	void (*SetPan)(int pan);	// from -127 to 127.. usually just call outMod->SetPan
	
	// in-window builtin vis stuff

	void (*SAVSAInit)(int maxlatency_in_ms, int srate);		// call once in Play(). maxlatency_in_ms should be the value returned from outMod->Open()
	// call after opening audio device with max latency in ms and samplerate
	void (*SAVSADeInit)();	// call in Stop()


	// simple vis supplying mode
	void (*SAAddPCMData)(void *PCMData, int nch, int bps, int timestamp); 
											// sets the spec data directly from PCM data
											// quick and easy way to get vis working :)
											// needs at least 576 samples :)

	// advanced vis supplying mode, only use if you're cool. Use SAAddPCMData for most stuff.
	int (*SAGetMode)();		// gets csa (the current type (4=ws,2=osc,1=spec))
							// use when calling SAAdd()
	void (*SAAdd)(void *data, int timestamp, int csa); // sets the spec data, filled in by winamp


	// vis stuff (plug-in)
	// simple vis supplying mode
	void (*VSAAddPCMData)(void *PCMData, int nch, int bps, int timestamp); // sets the vis data directly from PCM data
											// quick and easy way to get vis working :)
											// needs at least 576 samples :)

	// advanced vis supplying mode, only use if you're cool. Use VSAAddPCMData for most stuff.
	int (*VSAGetMode)(int *specNch, int *waveNch); // use to figure out what to give to VSAAdd
	void (*VSAAdd)(void *data, int timestamp); // filled in by winamp, called by plug-in


	// call this in Play() to tell the vis plug-ins the current output params. 
	void (*VSASetInfo)(int nch, int srate);


	// dsp plug-in processing: 
	// (filled in by winamp, called by input plug)

	// returns 1 if active (which means that the number of samples returned by dsp_dosamples
	// could be greater than went in.. Use it to estimate if you'll have enough room in the
	// output buffer
	int (*dsp_isactive)(); 

	// returns number of samples to output. This can be as much as twice numsamples. 
	// be sure to allocate enough buffer for samples, then.
	int (*dsp_dosamples)(short int *samples, int numsamples, int bps, int nch, int srate);


	// eq stuff
	void (*EQSet)(int on, char data[10], int preamp); // 0-64 each, 31 is +0, 0 is +12, 63 is -12. Do nothing to ignore.

	// info setting (filled in by winamp)
	void (*SetInfo)(int bitrate, int srate, int stereo, int synched); // if -1, changes ignored? :)

	Out_Module *outMod; // filled in by winamp, optionally used :)
} In_Module;

static void	 SetInfo(int, int, int, int);
static void	 EQSet(int, char[10], int);
static void	 VSASetInfo(int, int);
static void	 VSAAddPCMData(void *, int, int, int);
static int	 VSAGetMode(int *, int *);
static void	 VSAAdd(void *, int);
static void	 SAVSAInit(int, int);
static void	 SAVSADeInit();
static void	 SAAddPCMData(void *, int, int, int);
static int	 SAGetMode();
static void	 SAAdd(void *, int, int);
static int	 dsp_isactive();
static int	 dsp_dosamples(short int *, int, int, int, int);
static int	 Out_Open(int, int, int, int, int);
static void	 Out_Close();
static void	 Out_Flush(int);
static int 	 Out_Pause(int);
static int	 Out_Write(char *, int);
static int	 Out_CanWrite();
static int	 Out_IsPlaying();
static void	 Out_SetVolume(int);
static void	 Out_SetPan(int);
static int	 Out_GetOutputTime();
static int	 Out_GetWrittenTime();

static int bits;
static int channels;
static int rate;
static int running;
static int bitrate = 0;

int WINAPI WinMain(HINSTANCE eins, HINSTANCE zwei, LPSTR arg, int whatever)
{
	HINSTANCE dll;
	In_Module* module;
	char cmd[4096];
	
	if (!arg) {
		fprintf(stderr,"ERROR Usage: winamp <filename> <dll>\n");
		exit(1);
	}
	char *pipe = arg;
	while (*arg && *arg != ' ') arg++;
	if (!*arg) {
		fprintf(stderr,"ERROR Usage: winamp <filename> <dll>\n");
		exit(1);
	}
	*arg = 0;
	arg++;
	dll=LoadLibrary(arg);
	if (!dll) {
		sprintf(cmd,"/usr/local/lib/win32/%s",arg);
		dll=LoadLibrary(cmd);
	}
	if (!dll) {
		sprintf(cmd,"/usr/lib/win32/%s",arg);
		dll=LoadLibrary(cmd);
	}
	if (!dll) {
		fprintf(stderr,"ERROR could not load dll %s\n",arg);
		exit(1);
	}

	In_Module *(*proc)() = (In_Module *(*)()) GetProcAddress(dll, "winampGetInModule2");
	module = proc();
	module->hDllInstance=dll;
	module->SetInfo = SetInfo;
	module->EQSet = EQSet;
	module->VSASetInfo = VSASetInfo;
	module->VSAAddPCMData = VSAAddPCMData;
	module->VSAGetMode = VSAGetMode;
	module->VSAAdd = VSAAdd;
	module->SAVSAInit = SAVSAInit;
	module->SAVSADeInit = SAVSADeInit;
	module->SAAddPCMData = SAAddPCMData;
	module->SAGetMode = SAGetMode;
	module->SAAdd = SAAdd;
	module->dsp_isactive = dsp_isactive;
	module->dsp_dosamples = dsp_dosamples;
	module->outMod = calloc(sizeof(Out_Module),01);
	module->outMod->version = 256;
	module->outMod->description = "Winamp";
	module->outMod->id = 1;
	module->outMod->Open = Out_Open;
	module->outMod->Close = Out_Close;
	module->outMod->Write = Out_Write;
	module->outMod->CanWrite = Out_CanWrite;
	module->outMod->IsPlaying = Out_IsPlaying;
	module->outMod->Pause = Out_Pause;
	module->outMod->SetVolume = Out_SetVolume;
	module->outMod->SetPan = Out_SetPan;
	module->outMod->Flush = Out_Flush;
	module->outMod->GetOutputTime = Out_GetOutputTime;
	module->outMod->GetWrittenTime = Out_GetWrittenTime;

	module->Init();

	while (!feof(stdin)) {
		cmd[0] = 0;
		fgets(cmd,sizeof(cmd),stdin);
		if (!strcmp(cmd,"INIT\n")) {
			module->Stop();
			fprintf(stderr,"DLL %d %s\nOK\n",module->version,module->description);
		} else if (!strcmp(cmd,"QUIT\n")) {
			module->Stop();
		} else if (!strcmp(cmd,"START\n")) {
			module->Play(pipe);
		}
	}
	module->Stop();
	module->Quit();
	free(module->outMod);
	FreeLibrary(dll);
	
	return 0;
}


static void SetInfo(int br, int srate, int stereo, int synched)
{
	if (br != -1) bitrate = br;
}

static int Out_Open(int samplerate, int numchannels, int bitspersamp, int bufferlenms, int prebufferms)
{
	bits = bitspersamp;
	channels = numchannels;
	rate = samplerate;
	running = 1;
	fprintf(stderr,"STAT %i %i %i %i\n",bits/8,channels,rate,bitrate);
	return 100;
}

static int Out_IsPlaying() {
	running = 0;
	return 0;
}

static void Out_Close()
{
	fprintf(stderr,"QUIT\n");
}

static int Out_Write(char *buf, int len)
{
	write(1,buf,len);
	return 0;
}

static int Out_CanWrite() {
	return rate;
}

static void EQSet(int on, char data[10], int preamp) { }
static void VSASetInfo(int nch, int srate) { }
static void VSAAddPCMData(void *PCMData, int nch, int bps, int timestamp) { }
static int VSAGetMode(int *specNch, int *waveNch) { return 0; }
static void VSAAdd(void *data, int timestamp) { }
static void SAVSAInit(int latency, int srate) { }
static void SAVSADeInit() { }
static void SAAddPCMData(void *PCMData, int nch, int bps, int timestamp) { }
static int SAGetMode() { return 0; }
static void SAAdd(void *data, int timestamp, int csa) { }
static int dsp_isactive() { return 1; }
static int dsp_dosamples(short *s, int nsm, int bps, int nch, int srate) { return nsm; }
static int Out_Pause(int p) { return 0; }
static void Out_Flush(int t) { }
static int Out_GetOutputTime() { return 1; }
static int Out_GetWrittenTime() { return 1; }
static void Out_SetVolume(int vol) { }
static void Out_SetPan(int pan) { }
